function error_printer(msg, layer)
	--print((debug.traceback("Error: " .. tostring(msg), 1+(layer or 1)):gsub("\n[^\n]+$", "")))
end
 
function love.errhand(msg)
	print("* * * E R R O R * * *\n" .. msg)

	msg = tostring(msg)
 
	error_printer(msg, 2)
 
	if not love.window or not love.graphics or not love.event then
		return
	end
 
	-- Reset state.
	if love.mouse then
		love.mouse.setVisible(true)
		love.mouse.setGrabbed(false)
		if love.mouse.setRelativeMode ~= nil then
			love.mouse.setRelativeMode(false)
		end
	end
	if love.joystick then
		-- Stop all joystick vibrations.
		for i,v in ipairs(love.joystick.getJoysticks()) do
			v:setVibration()
		end
	end
	if love.audio then love.audio.stop() end
	
	-- I first want to make a screenshot of the current screen.
	local crashscreenshot = love.graphics.newImage(love.graphics.newScreenshot())

	love.graphics.reset()
	--local font = love.graphics.setNewFont(math.floor(love.window.toPixels(14)))
	
	font8 = love.graphics.newFont("Space Station.ttf", 8)
	font16 = love.graphics.newFont("Space Station.ttf", 16)
 
	--love.graphics.setBackgroundColor(89, 157, 220)
	love.graphics.setBackgroundColor(255, 0, 0)
	love.graphics.setColor(255, 255, 255, 255)
 
	--local trace = debug.traceback()
	local trace = ""
  
	love.graphics.clear(love.graphics.getBackgroundColor())
	love.graphics.origin()
 
	local err = {}
	
	if ctrl == nil then
		ctrl = "ctrl"
	end
	
	
	local mainmessage = msg:gsub("\n", ".") .. "\n\n" .. "    Ved version: a" .. ver .. "\n" .. "    LÖVE version: v0." .. (love.graphics.ellipse == nil and 9 or 10) .. "\n" .. "    State: " .. (state == nil and "nil" or state) .. "\n    OS: " .. love.system.getOS()
	
 
	table.insert(err, "Please tell Dav999 about this problem.\n\n\nDetails: (press " .. ctrl .. "+C to copy to the clipboard)\n\n")
	--[[
	table.insert(err, "    " .. msg:gsub("\n", ".") .. "\n") --.."\n\n")
	table.insert(err, "    Ved version: a" .. ver)
	table.insert(err, "    LÖVE version: v0." .. (love.graphics.ellipse == nil and 9 or 10))
	table.insert(err, "    State: " .. (state == nil and "nil" or state) .. "\n")
	]]
	table.insert(err, "    " .. mainmessage)
 
	for l in string.gmatch(trace, "(.-)\n") do
		if not string.match(l, "boot.lua") then
			l = string.gsub(l, "stack traceback:", "Traceback\n")
			table.insert(err, l)
		end
	end
 
	local p = table.concat(err, "\n")
 
	p = string.gsub(p, "\t", "")
	p = string.gsub(p, "%[string \"(.-)\"%]", "%1")
	
	
	
	if dialog == nil then
		--require("dialog")
		
		dialog = {}
		
		function dialog.draw() end
		function dialog.update() end
		function startmultiinput(_) end
	else
		-- We may have a dialog on the screen already. Remove that as to not have it running in the crash screen, because dialogs still work.
		DIAwindowani = 16
	end
	
	-- We want to make a dialog containing the crash. We don't, only for saving
	--dialog.new(p, "Fatal error", 1, 2, 0)

	local function draw()
		local pos = 40
		love.graphics.clear(love.graphics.getBackgroundColor())
		love.graphics.setColor(255,255,255,64)
		love.graphics.draw(crashscreenshot, 0, 0) --, 0, s.pscale^-1)
		
		-- Title
		love.graphics.setFont(font16)
		love.graphics.setColor(0,0,0,255)
		love.graphics.print("Ved has crashed!", pos+4, pos+4)
		love.graphics.setColor(255,255,255,255)
		love.graphics.print("Ved has crashed!", pos, pos)
		
		-- Draw a box for the important details
		love.graphics.setColor(255,92,92,208) -- 225 is gebruikt
		love.graphics.rectangle("fill", pos-2, pos+40+40-1, love.graphics.getWidth()-(2*pos)+4, 56+8)
		
		-- Main text
		love.graphics.setFont(font8)
		love.graphics.setColor(0,0,0,255)
		love.graphics.printf(p, pos+2, pos+40+2, love.graphics.getWidth() - pos + 2)
		love.graphics.setColor(255,255,255,255)
		love.graphics.printf(p, pos, pos+40, love.graphics.getWidth() - pos)
		
		dialog.draw()
		love.graphics.setColor(255,255,255,255)
		
		love.graphics.present()
	end
	local function update()
		dialog.update()
	end

	while true do
		love.event.pump()
 
		for e, a, b, c in love.event.poll() do
			if e == "quit" then
				return
			elseif e == "keypressed" and (a == "escape") then
				return
			elseif e == "keypressed" and a == "c" and (love.keyboard.isDown("l" .. ctrl) or love.keyboard.isDown("r" .. ctrl)) then
				love.system.setClipboardText(mainmessage:gsub("\n    ", "\n"))
			elseif e == "keypressed" and a == "s" then
				startmultiinput({(editingmap ~= "untitled\n" and editingmap or "")})
				dialog.new(L.ENTERNAMESAVE .. "\n\n\n\nNote: considering Ved has crashed, it may be better to save with a separate name now, to not overwrite your level in case the crash was caused by corruption in the level of some sorts.", "Unfortunately this doesn't work yet", 1, 4, 10)
			--[[
			elseif e == "mousepressed" and mousein(DIAx+DIAwidth-51, DIAy+DIAwindowani+DIAheight-26, DIAx+DIAwidth-1, DIAy+DIAwindowani+DIAheight-1) then
				dialog.push()
				DIAreturn = 1
				DIAquitting = 1
			]]
			elseif e == "touchpressed" then
				local name = love.window.getTitle()
				if #name == 0 or name == "Untitled" then name = "Game" end
				local buttons = {"OK", "Cancel"}
				local pressed = love.window.showMessageBox("Quit "..name.."?", "", buttons)
				if pressed == 1 then
					return
				end
			end
		end
 
		update()
		draw()
 
		if love.timer then
			love.timer.sleep(1/60)
		end
	end
 
end